// apachemod/SharedObject.hh
// This file is part of Anyterm; see http://anyterm.org/
// (C) 2005 Philip Endecott

// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.


// This file defines a namespace containing container types that use
// shared memory, using boost::shmem.  This implementation simplifies
// the raw boost::shmem interface by using single global shared memory
// segment.  Basically this code just removes to need to pass around
// so many extra template and constructor parameters.

#ifndef SharedObject_hh
#define SharedObject_hh

#include <boost/shmem/offset_ptr.hpp>
#include <boost/shmem/containers/map.hpp>
#include <boost/shmem/containers/vector.hpp>
#include <boost/shmem/containers/string.hpp>
#include <boost/shmem/containers/deque.hpp>
#include <boost/shmem/named_shared_object.hpp>
//#include <boost/shmem/allocators/node_allocator.hpp>
//#include <boost/shmem/allocators/private_node_allocator.hpp>
#include <boost/shmem/allocators/cached_node_allocator.hpp>

namespace SharedObject {

  // Set the shared memory segment to use:
  void set_segment(boost::shmem::named_shared_object& segment);

  // Segment manager for the single global shared memory segment
  extern boost::shmem::named_shared_object::segment_manager* segment_manager;

  // strings:

  template <typename CharT, typename TraitsT>
  struct basic_string:
    boost::shmem::basic_string<
      CharT,
      TraitsT,
      boost::shmem::cached_node_allocator<
        CharT,
	/*64,*/
        boost::shmem::named_shared_object::segment_manager
      >
    >
  {
    typedef boost::shmem::basic_string<
      CharT,
      TraitsT,
      boost::shmem::cached_node_allocator<
        CharT,
	/*64,*/
        boost::shmem::named_shared_object::segment_manager
      >
    > base_t;

    basic_string():
      base_t(boost::shmem::cached_node_allocator<
	       CharT,
	       /*64,*/
	       boost::shmem::named_shared_object::segment_manager
	     >(SharedObject::segment_manager))
    {}

    basic_string(const CharT* cstr):
      base_t(cstr,
	     boost::shmem::cached_node_allocator<
	       CharT,
    	       /*64,*/
	       boost::shmem::named_shared_object::segment_manager
	     >(SharedObject::segment_manager))
    {}

    basic_string(std::basic_string<CharT,TraitsT> str):
      base_t(str.c_str(),
	     boost::shmem::cached_node_allocator<
	       CharT,
	       /*64,*/
	       boost::shmem::named_shared_object::segment_manager
	     >(SharedObject::segment_manager))
    {}
  };

  typedef basic_string<char,std::char_traits<char> > string;


  // vectors:

  template <typename T>
  struct vector:
    boost::shmem::vector<
      T,
      boost::shmem::cached_node_allocator<
        T,
	/*64,*/
        boost::shmem::named_shared_object::segment_manager
      >
    >
  {
    typedef boost::shmem::vector<
      T,
      boost::shmem::cached_node_allocator<
        T,
	/*64,*/
        boost::shmem::named_shared_object::segment_manager
      >
    > base_t;

    vector():
      base_t(boost::shmem::cached_node_allocator<
	       T,
	       /*64,*/
	       boost::shmem::named_shared_object::segment_manager
	     >(segment_manager))
    {}

    vector(int n):
      base_t(n,
	     T(),
	     boost::shmem::cached_node_allocator<
	       T,
	       /*64,*/
               boost::shmem::named_shared_object::segment_manager
	     >(segment_manager))
    {}

  };


  // maps:

  template <typename K, typename V>
  struct map:
    boost::shmem::map<
      K,
      V,
      std::less<K>,
      boost::shmem::cached_node_allocator<
        std::pair<K,V>,
	/*64,*/
        boost::shmem::named_shared_object::segment_manager
      >
    >
  {
    typedef boost::shmem::map<
      K,
      V,
      std::less<K>,
      boost::shmem::cached_node_allocator<
        std::pair<K,V>,
	/*64,*/
        boost::shmem::named_shared_object::segment_manager
      >
    > base_t;

    map():
      base_t(std::less<K>(),
	     boost::shmem::cached_node_allocator<
	       std::pair<K,V>,
	       /*64,*/
	       boost::shmem::named_shared_object::segment_manager
	     >(segment_manager))
    {}

  };


  // deques:

  template <typename T>
  struct deque:
    boost::shmem::deque<
      T,
      boost::shmem::cached_node_allocator<
        T,
	/*64,*/
        boost::shmem::named_shared_object::segment_manager
      >
    >
  {
    typedef boost::shmem::deque<
      T,
      boost::shmem::cached_node_allocator<
        T,
	/*64,*/
        boost::shmem::named_shared_object::segment_manager
      >
    > base_t;

    deque():
      base_t(boost::shmem::cached_node_allocator<
	       T,
	       /*64,*/
	       boost::shmem::named_shared_object::segment_manager
	     >(segment_manager))
    {}

    deque(int n):
      base_t(n,
	     T(),
	     boost::shmem::cached_node_allocator<
	       T,
	       /*64,*/
               boost::shmem::named_shared_object::segment_manager
	     >(segment_manager))
    {}

  };

};

#endif

