// apachemod/backend.cc
// This file is part of Anyterm; see http://anyterm.org/
// (C) 2005-2006 Philip Endecott

// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.


#include "backend.h"

#include <sys/time.h>
#include <sys/resource.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>

#include <sstream>

#include "Iconver.hh"

#include "config.hh"
#include "anyterm.h"
#include "Exception.hh"
#include "Globals.hh"
#include "Terminal.hh"
#include "SubProcess.hh"

#include <boost/bind.hpp>

#include <boost/shmem/named_shared_object.hpp>


#define shm_size (2048*1024)  /* Size of shared memory region */

class Backend {
  // Private backend per-session state.
private:
  Session& ses;
  Terminal term;
  SubProcess process;
  void process_output(string s);
  void scrape_screen(void);

public:
  Backend(Session& ses_);
  void send_keys(string k) { process.send(k); }
};


void Backend::scrape_screen(void)
{
  // Copy the terminal's screen into the shared memory.

  for (int r=0; r<ses.rows; r++) {
    for (int c=0; c<ses.cols; c++) {
      RoteCell cell = ses.be->term.cell(r,c);
      ses.screen(c,r) = cell;
    }
  }
  ses.screen.cur_row = ses.be->term.cur_row();
  ses.screen.cur_col = ses.be->term.cur_col();
}


void Backend::process_output(string s)
{
  term.write(s);
  if (term.was_dirty()) {
    Mutex::scoped_lock l(ses.screen_lock);
    scrape_screen();
    ses.dirty = true;
    ses.dirty_condition.notify_all();
  }
}


static void report_backend_error(Session& ses, Exception& E)
{
  ostringstream s;
  E.report(s);
  ses.error_msg=s.str();
  ses.error=true;
}


Backend::Backend(Session& ses_):
  ses(ses_),
  term(ses.rows,ses.cols),
  process(boost::bind(&Backend::process_output,this,_1),
	  ses.command.c_str(),
	  ses.rows,ses.cols)
{}


static void backend_open (Session& ses)
{
  ses.be = new Backend(ses);
  ses.dirty=true;
  ses.error=false;
  ses.is_open=true;
}


static void backend_close (Session& ses)
{
  {
    Mutex::scoped_lock l(ses.screen_lock);
    ses.open_req=false;
    ses.is_open=false;
    ses.dirty=true;
    ses.dirty_condition.notify_all();
  }
  delete ses.be;
  ses.be=NULL; // flush out any bugs!
}


static void service_rote()
{
  // This is the loop for the backend process.  For each open session,
  // it receives keystrokes from the fifo, sends them to ROTE and
  // thence to the pty.  Opening new sessions and closing old ones
  // also happens along the way.

  // We'll be converting keyboard characters from UTF-8 to CHARSET:
  static Iconver utf8_to_CHARSET("UTF-8",CHARSET);

  Globals& globals(Globals::singleton());

  while (1) {

    {
      Mutex::scoped_lock l(globals.notification_mutex);
      if (!globals.notification) {
	struct boost::shmem::xtime timeout;
	xtime_get(&timeout, boost::shmem::TIME_UTC);
	timeout.sec += 10;
	globals.notification_condition.timed_wait(l,timeout);
      }
      globals.notification=false;
    }

    for (Globals::sessions_t::iterator i = globals.sessions.begin();
	 i != globals.sessions.end(); ++i) {

      Session& ses(i->second);

      try {
	
	if (ses.open_req && !ses.is_open) {
	  // Open new session if requested
	  backend_open(ses);
	}
	
	if (!ses.open_req && ses.is_open) {
	  // Close session if requested
	  backend_close(ses);
	}

	if (ses.is_open) {
	  // For open sessions:
	  // Check if timed out
	  
	  if ((time(NULL) - ses.last_access) > ses.time_out) {
	    backend_close(ses);
	    throw StrException("Session timed out");
	  }
	    
	  // Send keypresses
	  string keys;
	  while (!ses.key_fifo.empty()) {
	    keys += ses.key_fifo.front();
	    ses.key_fifo.pop();
	  }
	  string conv_keys = utf8_to_CHARSET(keys);
	  ses.be->send_keys(conv_keys);
	}
      }
      catch (Exception& E) {
	report_backend_error(ses,E);
      }

    }
  }
}


static void install_sigchld_handler(void);

static void reap_child(int sig)
{
  // Handler function for SIGCHLD in backend process
  // Note that apparently you might get only one SIGCHLD for multiple
  // subprocesses if they terminate at about the same time.  So you
  // can't just have
  //    wait(NULL);

  pid_t rc;
  do {
    rc = waitpid(-1, NULL, WNOHANG);
  } while (rc>0);

  install_sigchld_handler();
}

static void install_sigchld_handler(void)
{
  struct sigaction sa;
  sa.sa_handler=&reap_child;
  sigemptyset(&sa.sa_mask);
  sa.sa_flags=SA_NOCLDSTOP;
  sigaction(SIGCHLD,&sa,NULL);  // should check return value
}



static boost::shmem::named_shared_object shm_segment;

static void termination_handler(int sig)
{
  shm_segment.close();
  exit(0);
}


void backend_process(void)
{
  // It seems that Apache can put signal handling in an odd state with
  // various things ignored; this is not what our child processes
  // expect.  Sanitise signals.

  // These signal() calls can return errors, but not under any
  // realistic circumstances that I can think of.  Ignore them.

  for (int i=1; i<NSIG; i++) {
    signal(i,SIG_DFL);
  }
  sigset_t sigset;
  sigfillset(&sigset);
  sigprocmask(SIG_UNBLOCK,&sigset,NULL);


  // We don't want child processes to become zombies when they terminate.
  install_sigchld_handler();
  
  // Tidy up on termination
  signal(SIGTERM,&termination_handler);
  signal(SIGINT, &termination_handler);
  signal(SIGHUP, &termination_handler);


  // We will have inherited all sorts of file descriptors from Apache,
  // e.g. log files, sockets etc.  This doesn't do much harm until
  // they are inherited by a spawned process where they could even be
  // a security problem.  So just close them all now.
  struct rlimit files_limit;
  int rc = getrlimit(RLIMIT_NOFILE,&files_limit);
  if (rc==-1) {
    throw SysException("getrlimit(RLIMIT_NOFILE)");
  }
  int max_fds = files_limit.rlim_cur;
  for (int fd=0; fd<max_fds; fd++) {
    close(fd);
  }

  // Create the shared memory:
  bool ok = shm_segment.create(shm_name, shm_size);
  if (!ok) {
    throw SysException("shm_segment.create()");
  }
  SharedObject::set_segment(shm_segment);
  Globals::set_singleton(shm_segment.find_or_construct<Globals>("globals")());

  // Start updating the terminals
  service_rote();
}

